--
-- Database creation
--

CREATE EXTENSION IF NOT EXISTS dblink;

DO $$
DECLARE
    db_name TEXT;
BEGIN

    db_name := 'pscheduler';

    IF NOT EXISTS (SELECT * FROM pg_roles WHERE rolname = db_name)
    THEN
        PERFORM dblink_exec('dbname=' || current_database(),
                            'CREATE ROLE pscheduler WITH LOGIN');
        RAISE NOTICE 'Created role %', db_name;
    END IF;

    IF NOT EXISTS (SELECT * FROM pg_database WHERE datname = db_name)
    THEN
        PERFORM dblink_exec('dbname=' || current_database(),
                            'CREATE DATABASE pscheduler WITH OWNER=pscheduler');
        PERFORM dblink_exec('dbname=' || current_database(),
                            'GRANT ALL ON DATABASE pscheduler TO pscheduler');
        RAISE NOTICE 'Created database %', db_name;
    END IF;

END;
$$ LANGUAGE plpgsql;


\c pscheduler

CREATE OR REPLACE LANGUAGE plpythonu;

-- Needed for generating UUIDs
CREATE EXTENSION IF NOT EXISTS pgcrypto;
--
-- Utilities required for the superuser and others
--

-- ----------------------------------------------------------------------------

--
-- Database Manipulation
--

-- Drop all variations of a function with cascading.
-- Adapted from https://stackoverflow.com/a/7622951/180674

CREATE OR REPLACE FUNCTION drop_function_all(functionname text)
RETURNS VOID AS
$$
DECLARE
    funcrow RECORD;
    numparameters INTEGER;
    param INTEGER;
    paramtext TEXT;
BEGIN
    FOR funcrow IN
        SELECT proargtypes FROM pg_proc WHERE proname = functionname
    LOOP
    
        -- For some reason array_upper is off by one for the oidvector
        -- type, hence the +1
        numparameters = array_upper(funcrow.proargtypes, 1) + 1;
    
        param := 0;
        paramtext := '';
    
        LOOP
            IF param < numparameters THEN
                IF param > 0 THEN
                    paramtext = paramtext || ', ';
                END IF;
                paramtext := paramtext || (
		    SELECT typname FROM pg_type
		    WHERE oid = funcrow.proargtypes[param]
		);
                param := param + 1;
            ELSE
                EXIT;
            END IF;
        END LOOP;
	    
        EXECUTE 'DROP FUNCTION ' || functionname || '(' || paramtext 
	    || ') CASCADE;';
    
    END LOOP;

END;
$$
LANGUAGE plpgsql VOLATILE;
--
-- Stored procedure that runs external programs
--

-- NOTE: This uses an untrusted language and therefore must be
-- installed as the superuser

DROP TYPE IF EXISTS external_program_result CASCADE;
CREATE TYPE external_program_result AS (
       status INTEGER,
       stdout TEXT,
       stderr TEXT
);


--
-- Run a pscheduler frontend command
--
-- TODO: Figure out what do do when DB is off-node.  Probably need an
-- external process like the task runner.
--
-- TODO: This doesn't fail if the program never returns.  Need to do
-- something about that, maybe along the lines of
-- http://stackoverflow.com/a/4825933.  (Uses threading... Ick.)
--

DO $$ BEGIN PERFORM drop_function_all('pscheduler_command'); END $$;

CREATE OR REPLACE FUNCTION pscheduler_command(
    argv TEXT[] DEFAULT '{}',  -- Arguments to pass to 'pscheduler'
    input TEXT DEFAULT NULL,   -- What goes to the standard input
    timeout INTEGER DEFAULT 5  -- How long to wait in seconds
)
RETURNS external_program_result
AS $$

import pscheduler

try:

    status, stdout, stderr = pscheduler.run_program(
        [ 'pscheduler' ] + argv,
	stdin = input,
	timeout = timeout
	)

except Exception as ex:
    plpy.error('Unable to run pscheduler front-end program: ' + str(ex))

return { 'status': status, 'stdout': stdout, 'stderr': stderr }

$$ LANGUAGE plpythonu;
--
-- Stored procedures for HTTP(S) Operations
--

-- NOTE: This uses an untrusted language and therefore must be
-- installed as the superuser

-- TODO: It would be nice to use PgCURL
-- (https://github.com/Ormod/pgcurl) or something like it instead of
-- dragging this through Python.  This'll do for now.

DROP TYPE IF EXISTS http_result CASCADE;
CREATE TYPE http_result AS (
       status INTEGER,
       returned TEXT
);


DO $$ BEGIN PERFORM drop_function_all('http_get'); END $$;

CREATE OR REPLACE FUNCTION http_get(
    url TEXT DEFAULT NULL,
    timeout FLOAT default NULL,
    bind TEXT default NULL
)
RETURNS http_result
AS $$

import pscheduler

try:
    status, returned = pscheduler.url_get(url, json=False, throw=False,
        timeout=timeout, bind=bind)
except Exception as ex:
    plpy.error("Failed to GET %s: %s" % (url, str(ex)))

return { "status": status, "returned": returned }

$$ LANGUAGE plpythonu;


DO $$ BEGIN PERFORM drop_function_all('http_put'); END $$;

CREATE OR REPLACE FUNCTION http_put(
    url TEXT DEFAULT NULL,
    payload TEXT DEFAULT NULL,
    timeout FLOAT default NULL,
    bind TEXT default NULL
)
RETURNS http_result
AS $$

import pscheduler

try:
    status, returned = pscheduler.url_put(url, data=payload,
                                          json=False, throw=False,
                                          timeout=timeout, bind=bind)
except Exception as ex:
    plpy.error("Failed to PUT %s: %s" % (url, str(ex)))

return { "status": status, "returned": returned }

$$ LANGUAGE plpythonu;


DO $$ BEGIN PERFORM drop_function_all('http_post'); END $$;

CREATE OR REPLACE FUNCTION http_post(
    url TEXT DEFAULT NULL,
    payload TEXT DEFAULT NULL,
    timeout FLOAT default NULL,
    bind TEXT default NULL
)
RETURNS http_result
AS $$

import pscheduler

try:
    status, returned = pscheduler.url_post(url, data=payload,
                                           json=False, throw=False,
                                           timeout=timeout, bind=bind)
except Exception as ex:
    plpy.error("Failed to POST %s: %s" % (url, str(ex)))

return { "status": status, "returned": returned }

$$ LANGUAGE plpythonu;


DO $$ BEGIN PERFORM drop_function_all('http_delete'); END $$;

CREATE OR REPLACE FUNCTION http_delete(
    url TEXT DEFAULT NULL,
    timeout FLOAT default NULL,
    bind TEXT default NULL
)
RETURNS http_result
AS $$

import pscheduler

try:
    status, returned = pscheduler.url_delete(url, throw=False,
    	    timeout=timeout, bind=bind)
except Exception as ex:
    plpy.error("Failed to DELETE %s: %s" % (url, str(ex)))

return { "status": status, "returned": returned }

$$ LANGUAGE plpythonu;
--
-- Stored procedure that validates JSON using the pScheduler library
--

-- NOTE: This uses an untrusted language and therefore must be
-- installed as the superuser

--
-- Validate JSON
--
--

DO $$ BEGIN PERFORM drop_function_all('json_validate'); END $$;

CREATE OR REPLACE FUNCTION json_validate(
    candidate JSONB,                               -- JSON to be validated
    json_ref TEXT DEFAULT '#/pScheduler/AnyJSON'   -- What object to validate
)
RETURNS TEXT  -- NULL if okay, error text otherwise.
AS $$

import pscheduler

if not isinstance(candidate, basestring):
    plpy.error("No idea what to do with type %s" % (type(candidate)))

json = pscheduler.json_load(candidate)

valid, message = pscheduler.json_validate(
    {"": json},
    {    
        "type": "object",
        "properties": {
            "": { "$ref": json_ref }
        },
        "required": [ "" ]
    })

return None if valid else message

$$ LANGUAGE plpythonu;
